@@ -14,7 +14,8 @@ import numpy as np
 #                            #
 # ========================== #
 
-KNOWN_SCIENCE_SUFFIXES = {    "sciimg.fits":"(primary science image)",
+KNOWN_SCIENCE_SUFFIXES = {
+                      "sciimg.fits":"(primary science image)",
                       "mskimg.fits":"(bit-mask image)",
                       "psfcat.fits":"(PSF-fit photometry catalog)",
                       "sexcat.fits":"(nested-aperture photometry catalog)",
@@ -30,9 +31,101 @@ KNOWN_SCIENCE_SUFFIXES = {    "sciimg.fits":"(primary science image)",
 DATA_BASEURL   = "https://irsa.ipac.caltech.edu/ibe/data/ztf/products/"
 ZTFIRSA_BASE = ["/ref/","/sci/","/raw/","/cal/"]
 
-from .io import LOCALSOURCE
+
+from .io import LOCALSOURCE, CCIN2P3_SOURCE
 DEFAULT_SOURCE = DATA_BASEURL
 
+FILTERS = {"zg":1,"zr":2,"zi":3, "OO":None}
+
+
+# ================== #
+#  Filename parsing  #
+# ================== #
+def parse_filename(filename):
+    """ """
+    kind = filename_to_kind(filename)
+    if kind == "cal":
+        return parse_calfilename(filename)
+    if kind == "raw":
+        return parse_rawfilename(filename)
+    if kind == "sci":
+        return parse_scifilename(filename)
+
+    raise ValueError(f"Cannot parse the file {filename}")
+
+def filename_to_kind(filename):
+    """ """
+    filesplit = os.path.basename(filename).split("_")
+    if len(filesplit) == 6:
+        if "_hifreq" in filename or "_bias" in filename:
+            return "cal"
+        else:
+            return "raw"
+        
+    elif len(filesplit)>=8:
+        return "sci"
+
+    raise ValueError(f"Cannot parse the file {filename} ; remark 'ref' not implemented.")
+
+def parse_calfilename(filename):
+    """ """
+    _, fileday, filtercode, ccd_, qid_, suffix = os.path.basename(filename).split("_")
+    year,month, day = fileday[:4],fileday[4:6], fileday[6:]
+    ccdid = int(ccd_.replace("c",""))
+    qid = int(qid_.replace("q",""))
+    return {"year":year,
+            "month":month,
+            "day":day,
+            "fileday":fileday,
+            "ccdid":ccdid,
+            "qid":qid,
+            "rcid":ccdid_qid_to_rcid(ccdid,qid),
+            "filtercode":filtercode,
+            "filterid":FILTERS[filtercode],
+            "kind":"cal",
+            "type":suffix.split(".")[0]}
+
+    
+def parse_rawfilename(filename):
+    """ """
+    _, filefracday, paddedfield, filtercode, ccd_, suffix_ = os.path.basename(filename).split("_")
+    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
+    ccdid = int(ccd_.replace("c",""))
+    return {"year":year,
+            "month":month,
+            "day":day,
+            "filefracday":filefracday,
+            "paddedfield":paddedfield,
+            "field":int(paddedfield),
+            "ccdid":ccdid,
+            "filtercode":filtercode,
+            "filterid":FILTERS[filtercode],
+            "kind":"raw",
+            "suffix":suffix_}
+
+    
+def parse_scifilename(filename):
+    """ """
+    from .fields import ccdid_qid_to_rcid
+    
+    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, *suffix_ = os.path.basename(filename).split("_")
+    year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
+
+    ccdid = int(ccd_.replace("c",""))
+    qid = int(qid_.replace("q",""))
+    return {"year":year,
+            "month":month,
+            "day":day,
+            "filefracday":filefracday,
+            "paddedfield":paddedfield,
+            "field":int(paddedfield),
+            "ccdid":ccdid,
+            "qid":qid,
+            "rcid":ccdid_qid_to_rcid(ccdid,qid),
+            "filtercode":filtercode,
+            "filterid":FILTERS[filtercode],
+            "kind":"sci",
+            "suffix":"_".join(suffix_)}
 
 # ================== #
 #  Building the URL  #
@@ -66,6 +159,9 @@ def _source_to_location_(source):
     if source in ["local", "cmp", "computer"]:
         return LOCALSOURCE
     
+    if source in ["cc", "ccin2p3"]:
+        return CCIN2P3_SOURCE
+    
     if source in ['nersc']:
         raise ValueError("NERCS sourcing not ready yet.")
     
@@ -79,7 +175,7 @@ def science_path(year, month, day, fracday,
                 paddedfield,
                 filtercode, paddedccdid, qid,
                 imgtypecode="o", suffix="sciimg.fits",
-                source="", verbose=False):
+                source="", verbose=False, check_suffix=True):
     """ 
     suffix: [string]
         What kind of data do you want?
@@ -106,12 +202,12 @@ def science_path(year, month, day, fracday,
     
     """
     if verbose:
-        print(locals())
+        print(f"science_path: {locals()}")
     source = _source_to_location_(source)
     if suffix is None:
         suffix="sciimg.fits"
-    elif suffix not in KNOWN_SCIENCE_SUFFIXES.keys():
-        raise ValueError("Unkwown suffix %s for 'sci' exposures"%suffix, "\n known suffixes: \n", KNOWN_SCIENCE_SUFFIXES)
+    elif suffix not in KNOWN_SCIENCE_SUFFIXES.keys() and check_suffix:
+        raise ValueError(f"Unkwown suffix {suffix} for 'sci' exposures", "\n known suffixes: \n", KNOWN_SCIENCE_SUFFIXES)
     
     filefracday = "".join([year+month+day+fracday])
     file_ = 'sci/'+year+'/'+month+day+'/'+fracday+'/ztf_'+filefracday+'_'+paddedfield+'_'+filtercode+'_c'+paddedccdid+'_'+imgtypecode+'_q'+qid+'_'+suffix
@@ -151,7 +247,7 @@ def calibration_path(caltype,
                     year, month, day,
                     filtercode,
                     paddedccdid, qid, suffix=None,
-                    source=""):
+                    source="", check_suffix=True):
     """ 
     Parameters:
     -----------
@@ -178,8 +274,8 @@ def calibration_path(caltype,
     # ================ #
     source = _source_to_location_(source)
     caltype = caltype.lower() # case insensitive
-    if caltype.lower() not in ["bias", "hifreqflat"]:
-        raise ValueError("Unknown `caltype` %s"%caltype+" caltype must be  'bias' or 'hifreqflat'")
+    if caltype.lower() not in ["bias", "hifreqflat"] and check_suffix:
+        raise ValueError(f"Unknown `caltype` {caltype}, caltype must be  'bias' or 'hifreqflat'")
 
     if suffix is None or suffix in ["basic",".fits","fits"]:
         suffix = caltype+".fits"
@@ -229,7 +325,7 @@ def reference_path(paddedfield,
 
     """
     if verbose:
-        print(locals())
+        print(f"reference_path: {locals()}")
 
     if suffix is None:
         suffix = "refimg.fits"
@@ -240,10 +336,47 @@ def reference_path(paddedfield,
 # ============= #
 #   TOOLS       #
 # ============= #
-def filename_to_scienceurl(filename, suffix=None, source="irsa", verbose=False):
+def build_filename_from_dataframe(dataframe, suffix="sciimg.fits"):
+    """ The dataframe must contains:
+    filefracday, fieldid, ccdid, qid and filterid
+
+    Returns
+    ------
+    pandas.Series (name of the files (not fullpath))
+    """
+    filefrac    = dataframe["filefracday"].astype("int").astype("str")
+    paddedfield = dataframe["fieldid"].astype("int").astype("str").str.pad(6, fillchar="0")
+    paddedccd = "c"+dataframe["ccdid"].astype("int").astype("str").str.pad(2, fillchar="0")
+    paddedqid = "q"+dataframe["qid"].astype("int").astype("str")
+    filtername = dataframe["filterid"].apply(lambda x: "zg" if x==1 else "zr" if x==2 else "zi")
+    return "ztf"+"_"+filefrac+"_"+paddedfield+"_"+filtername+"_"+paddedccd+"_"+"o"+"_"+paddedqid+"_"+suffix
+
+
+def filename_to_url(filename, suffix=None, source="irsa", verbose=False, check_suffix=True):
+    """ Generic high level function that first detects the inputfile kind and then 
+    calls the associated function ; e.g., filename_to_scienceurl or filename_to_calurl.
+
+    NB:
+    suffix -> imgtypecode for raw images.
+    """
+    kind = filename_to_kind(filename)
+    if kind == "sci":
+        return filename_to_scienceurl(filename, suffix=suffix, source=source, verbose=verbose, check_suffix=check_suffix)
+    
+    if kind == "raw":
+        return filename_to_rawurl(filename, imgtypecode=suffix, source=source, verbose=verbose, check_suffix=check_suffix)
+    
+    if kind == "cal":
+        return filename_to_calurl(filename, suffix=suffix, source=source, verbose=verbose, check_suffix=check_suffix)
+
+    raise ValueError(f"Cannot parse the file {filename}")
+    
+def filename_to_scienceurl(filename, suffix=None, source="irsa", verbose=False, check_suffix=True):
     """ 
     """
-    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, suffix_ = os.path.basename(filename).split("_")
+    _, filefracday, paddedfield, filtercode, ccd_, imgtypecode, qid_, *suffix_ = os.path.basename(filename).split("_")
+    suffix_ = "_".join(suffix_)
+    
     year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
     paddedccdid = ccd_.replace("c","")
     qid = qid_.replace("q","")
@@ -255,8 +388,42 @@ def filename_to_scienceurl(filename, suffix=None, source="irsa", verbose=False):
                         paddedfield, filtercode,
                         paddedccdid, qid, # added in input
                         imgtypecode=imgtypecode, suffix=suffix,
-                        source=source, verbose=verbose)
+                        source=source, verbose=verbose,
+                        check_suffix=check_suffix)
     
+def filename_to_calurl(filename, suffix=None, source="irsa", verbose=False, check_suffix=True):
+    """ 
+    """
+    _, date, filtercode, ccd_, qid_, *suffix_ = os.path.basename(filename).split("_")
+    year,month,day = date[:4],date[4:6],date[6:]
+    suffix_ = "_".join(suffix_)
+    caltype = suffix_.split(".")[0]
+    paddedccdid = ccd_[1:]
+    qid = qid_[1:]
+    if suffix is None:
+        suffix = suffix_
+    
+    return calibration_path(caltype,
+                                year, month, day,
+                                filtercode,
+                                paddedccdid, qid, suffix=suffix,
+                                source=source, check_suffix=True)
+
+
+def filename_to_rawurl(filename, imgtypecode=None, source="irsa", verbose=False, check_suffix=True):
+    """ 
+    """
+    _, filefracday, paddedfield ,filtercode, ccd_, suffix_ =  os.path.basename(filename).split("_")
+    year, month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
+    imgtypecode_ = suffix_.split(".")[0]
+    paddedccdid = ccd_[1:]
+    if imgtypecode is None:
+        imgtypecode = imgtypecode_
+        
+    return raw_path(year, month, day, fracday,
+                             paddedfield, filtercode, paddedccdid, imgtypecode=imgtypecode,
+                             source=source)
+
 
 def filename_to_refurl(filename, suffix, source="irsa", verbose=False):
     """ 
@@ -275,7 +442,6 @@ def filename_to_refurl(filename, suffix, source="irsa", verbose=False):
     paddedccdid = ccd_.replace("c","")
     qid = qid_.replace("q","")
     
-    
     return reference_path(paddedfield,
                                    filtercode,
                                    paddedccdid, qid,
@@ -292,22 +458,19 @@ def filefrac_to_year_monthday_fracday(filefracday):
 
 def fileroot_to_science_url(fileroot, paddedccdid, qid,
                             imgtypecode="o", suffix="sciimg.fits", source="",
-                            verbose=False):
+                            verbose=False, check_suffix=True):
     """ 
     Parameters
     ----------
 
     """
-    if suffix not in KNOWN_SCIENCE_SUFFIXES.keys():
-        raise ValueError("Unkwown suffix %s for 'sci' exposures"%suffix , "\n known suffixes: \n", KNOWN_SCIENCE_SUFFIXES)
-    
     _, filefracday, paddedfield, filtercode = fileroot.split("_")
     year,month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
     return science_path(year, month, day, fracday,
                 paddedfield, filtercode,
                 paddedccdid, qid, # added in input
                 imgtypecode=imgtypecode, suffix=suffix,
-                source=source, verbose=verbose)
+                source=source, verbose=verbose, check_suffix=check_suffix)
 
 
 def fileroot_to_raw_url(fileroot, paddedccdid,
