@@ -16,7 +16,7 @@ from . import io
 
 
 # Combining metadata with buildurl
-def metatable_to_url(metatable, datakind='sci', suffix=None, source=None):
+def metatable_to_url(metatable, datakind='sci', suffix=None, source=None, **kwargs):
     """generic method to build the url/fullpath or the requested data.
         This method is based on the `builurl.py` module of ztfquery.
         
@@ -75,7 +75,7 @@ def metatable_to_url(metatable, datakind='sci', suffix=None, source=None):
                 # LIST of URL to download [SCIENCE]
             return  [buildurl.science_path(year_, month_, day_, fracday_, paddedfield_,
                                 filtercode_, paddedccdid_, qid_,
-                                imgtypecode=imgtypecode_, suffix=suffix, source=source)
+                                imgtypecode=imgtypecode_, suffix=suffix, source=source, **kwargs)
                         for year_, month_, day_, fracday_, paddedfield_, filtercode_,
                         paddedccdid_, qid_, imgtypecode_
                         in zip(year, month, day, fracday, paddedfield, filtercode,
@@ -85,7 +85,7 @@ def metatable_to_url(metatable, datakind='sci', suffix=None, source=None):
             return  [buildurl.raw_path(year_, month_, day_, fracday_,
                                            paddedfield_ if imgtypecode_ != "f" else '000000',  # because sometime they do have a field, why is that ?,
                               filtercode_, paddedccdid_, 
-                              imgtypecode=imgtypecode_, source=source)
+                              imgtypecode=imgtypecode_, source=source, **kwargs)
                         for year_, month_, day_, fracday_, paddedfield_, filtercode_,
                         paddedccdid_,  imgtypecode_
                         in zip(year, month, day, fracday, paddedfield, filtercode,
@@ -104,7 +104,7 @@ def metatable_to_url(metatable, datakind='sci', suffix=None, source=None):
         return  [buildurl.calibration_path(caltype_,
                                             year_, month_, day_,
                                             filtercode_, paddedccdid_, qid_,
-                                            suffix=suffix, source=source)
+                                            suffix=suffix, source=source, **kwargs)
                             for caltype_, year_, month_, day_, filtercode_, paddedccdid_, qid_
                      in zip(caltype,year, month, day,filtercode, paddedccdid, qid) ]
     # PIXELS
@@ -117,7 +117,7 @@ def metatable_to_url(metatable, datakind='sci', suffix=None, source=None):
                                 filtercode_, paddedccdid_, qid_,
                                 suffix=suffix,
                                 fieldprefix=paddedfield_[:3], # This is how it is defined in IRSA
-                                source=source)
+                                source=source, **kwargs)
                         for  paddedfield_, filtercode_, paddedccdid_, qid_
                     in zip(paddedfield, filtercode, paddedccdid, qid)]
         
@@ -156,12 +156,10 @@ class _ZTFDownloader_( object ):
     
     # Generic that should automatically work as long as get_data_path is defined.
     def download_data(self, suffix=None, source="IRSA", indexes=None,
-                     download_dir=None,
-                     show_progress = True, notebook=False, verbose=True,
-                     nodl = False, overwrite=False, nprocess=None,
-                     auth=None,
+                     download_dir=None, show_progress = True, verbose=True,
+                     nodl = False, overwrite=False, nprocess=None, auth=None,
                      filecheck=True, erasebad=True, redownload=True,
-                     ignore_warnings=False,
+                     ignore_warnings=False, client=None, 
                      **kwargs):
         """ 
         Parameters
@@ -237,7 +235,8 @@ class _ZTFDownloader_( object ):
             cookie = None
             
         # Data Structure
-        self._relative_data_path = self.get_data_path(suffix=suffix, source="None", indexes=indexes, **kwargs)
+        self._relative_data_path = self.get_data_path(suffix=suffix, source="None",
+                                                      indexes=indexes, **kwargs)
         
         # The IRSA location
         self.to_download_urls    = [os.path.join(buildurl._source_to_location_(source),d_)
@@ -257,11 +256,14 @@ class _ZTFDownloader_( object ):
         with warnings.catch_warnings():
             if ignore_warnings:
                 warnings.simplefilter("ignore")
-            io.download_url(self.to_download_urls, self.download_location,
-                        show_progress = show_progress, notebook=notebook, verbose=verbose,
+            delayed_ = io.download_url(self.to_download_urls, self.download_location,
+                        show_progress = show_progress, verbose=verbose,
                         overwrite=overwrite, nprocess=nprocess, cookies=cookie,
-                        filecheck=False)
-
+                        filecheck=False, client=client)
+            print("delayed")
+            if client is not None:
+                return delayed_
+            
         if filecheck:
             with warnings.catch_warnings():
                 if ignore_warnings:
@@ -276,9 +278,31 @@ class _ZTFDownloader_( object ):
     # --------- #
     #  GETTER   #
     # --------- #
+    def get_data(self, suffix=None, indexes=None, downloadit=True, check_suffix=True,
+                     dlfrom="irsa", overwrite=False, maxnprocess=4, exist=True,
+                     show_progress=True, **kwargs):
+        """ high-level method to get the data you want.
+        It will look for the data locally and download the one that are missing.
+        Set exist = False to have the expected list of data.
+
+        suffix could be a list of suffix, e.g. ["sciimg.fits", "mskimg.fits"]
+        
+        Parameters
+        ----------
+
+        Returns
+        -------
+        list of path.
+        """
+        return io.get_file(self.get_data_path(indexes=indexes), suffix=suffix, exist=exist,
+                               downloadit=downloadit, show_progress=show_progress, dlfrom=dlfrom,
+                               maxnprocess=maxnprocess, check_suffix=check_suffix,
+                               overwrite=overwrite, **kwargs)
+    
+    
     def get_local_data(self, suffix=None, exists=True, filecheck=True, indexes=None,
                            badfiles=False, source="local",
-                           ignore_warnings=False, **kwargs):
+                           ignore_warnings=False, fileout=None, check_suffix=True, **kwargs):
         """ the lists of files stored in your local copy of the ztf database.
         [This methods uses the get_data_path() method assuming source='local']
 
@@ -332,7 +356,7 @@ class _ZTFDownloader_( object ):
         -------
         list
         """
-        files = self.get_data_path(suffix=suffix, source=source, indexes=indexes)
+        files = self.get_data_path(suffix=suffix, source=source, indexes=indexes, check_suffix=check_suffix)
         if not exists:
             return files
 
@@ -348,7 +372,11 @@ class _ZTFDownloader_( object ):
                 return badfiles_
             elif badfiles_ is not None and len(badfiles_)>1:
                 return [f for f in localfile if f not in badfiles_]
-            
+
+        if fileout is not None:
+            with open(fileout, "w") as f_:
+                [f_.write(l+"\n") for l in localfile]
+                
         return localfile
 
     def get_missing_data_index(self, suffix=None, which="any"):
@@ -456,7 +484,8 @@ class _ZTFDownloader_( object ):
         """
         with warnings.catch_warnings():
             warnings.simplefilter("ignore")
-            badfiles = self.get_local_data(suffix=suffix, exists=True, filecheck=True, indexes=indexes, badfiles=True,
+            badfiles = self.get_local_data(suffix=suffix, exists=True, filecheck=True,
+                                            indexes=indexes, badfiles=True,
                                             redownload=redownload, nprocess=nprocess, **kwargs)
 
         if badfiles is None or len(badfiles)==0:
@@ -497,14 +526,15 @@ class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
                            sql_query=None, auth=None, **kwargs):
         """ Loads a ZTFQuery instance and runs load_metadata with the given input """
         this = cls()
-        this.load_metadata(kind=kind, radec=radec, size=size, caltype=caltype, sql_query=sql_query, auth=auth, **kwargs)
+        this.load_metadata(kind=kind, radec=radec, size=size, caltype=caltype,
+                               sql_query=sql_query, auth=auth, **kwargs)
         return this
     
     def set_metatable(self, metatable, kind):
         """ directly provide the metatable of interest"""
         self._metatable = metatable
-        if kind not in ["sci","raw","calib","ref"]:
-            raise ValueError(f"The given metatable should be of kind sci/raw/calib/ref, {kind} given")
+        if kind not in ["sci","raw","cal","ref"]:
+            raise ValueError(f"The given metatable should be of kind sci/raw/cal/ref, {kind} given")
         
         self._datakind = kind
         
@@ -515,7 +545,8 @@ class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
                         radec=None, size=None, mcen=None,
                         caltype=None,
                         sql_query=None, auth=None, **kwargs):
-        """ Querying for the metadata information that enables to reconstruct the URL to access the data.
+        """ Querying for the metadata information that enables to reconstruct 
+        the URL to access the data.
         
         [This methods uses the .metasearch library, which is python wrapper of the the IRSA web API
         see https://irsa.ipac.caltech.edu/docs/program_interface/ztf_api.html]
@@ -535,10 +566,12 @@ class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
         sql_query: [None or string] -optional - 
             The where parameter can be set to a 'SQL WHERE' clause, with some restrictions. 
             [https://en.wikipedia.org/wiki/Where_(SQL)]
-            Notably, function calls and sub-queries are not supported. You can use AND, OR, NOT, IN, BETWEEN, LIKE, IS, 
+            Notably, function calls and sub-queries are not supported. 
+            You can use AND, OR, NOT, IN, BETWEEN, LIKE, IS, 
             the usual arithmetic and comparison operators, and literal values.
             Note that the where parameter is required in the absence of POS (a spatial constraint).
-            WHERE clauses should be URL encoded [https://en.wikipedia.org/wiki/Query_string#URL_encoding].
+            WHERE clauses should be URL encoded 
+            [https://en.wikipedia.org/wiki/Query_string#URL_encoding].
             for instance  SPACE is encoded as '+' or "%20".
             If entry must be equal to a string, use `entry='str'` (with the quotes)
             Examples:
@@ -600,7 +633,8 @@ class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
         
         if kind not in ['cal']:
             # python3 -> self.metaquery = download_metadata(**{**locals(),**kwargs})
-            self.metaquery = download_metadata(kind=kind, radec=radec, size=size, mcen=mcen, sql_query=sql_query, **kwargs)
+            self.metaquery = download_metadata(kind=kind, radec=radec, size=size,
+                                                   mcen=mcen, sql_query=sql_query, **kwargs)
         else:
             for k in ["radec", "size", "mcen"]:
                 if locals()[k] is not None: warnings.warn("You are querying 'calibration data', the following entry is ignored: %s"%k)
@@ -610,7 +644,7 @@ class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
             self.metaquery = download_metadata(kind=kind, sql_query=sql_query, **kwargs)
             
     
-    def get_data_path(self, suffix=None, source=None, indexes=None):
+    def get_data_path(self, suffix=None, source=None, indexes=None, **kwargs):
         """ generic method to build the url/fullpath or the requested data.
         This method is based on the `builurl.py` module of ztfquery.
         
@@ -655,8 +689,11 @@ class ZTFQuery( ztftable._ZTFTable_, _ZTFDownloader_ ):
         if len(self.metatable) == 0:
             warnings.warn("No entry associated to the query you made: metatable is empty")
             return []
+        if "check_suffix" in kwargs and self.datakind not in ["cal","sci"]:
+            _ = kwargs.pop("check_suffix")
+            
         return metatable_to_url(self.metatable if indexes is None else self.metatable.loc[np.atleast_1d(indexes)],
-                                    datakind=self.datakind, suffix=suffix, source=source)
+                                    datakind=self.datakind, suffix=suffix, source=source, **kwargs)
 
     def get_local_metatable(self, suffix=None, which="any", invert=False):
         """ 
